@@ -1,7 +1,8 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 import datetime
-from sql import Literal
+
+from sql import Literal, Null
 from sql.aggregate import Count
 
 from trytond.model import ModelView, ModelSQL, MatchMixin, fields
@@ -10,6 +11,8 @@ from trytond.pool import Pool, PoolMeta
 from trytond.transaction import Transaction
 from trytond import backend
 
+from trytond.modules.product import price_digits
+
 __all__ = ['Template', 'Product', 'ProductSupplier', 'ProductSupplierPrice']
 __metaclass__ = PoolMeta
 
@@ -54,18 +57,22 @@ class Template:
     @fields.depends('default_uom', 'purchase_uom', 'purchasable')
     def on_change_default_uom(self):
         try:
-            changes = super(Template, self).on_change_default_uom()
+            super(Template, self).on_change_default_uom()
         except AttributeError:
-            changes = {}
+            pass
         if self.default_uom:
             if self.purchase_uom:
-                if self.default_uom.category == self.purchase_uom.category:
-                    changes['purchase_uom'] = self.purchase_uom.id
-                else:
-                    changes['purchase_uom'] = self.default_uom.id
+                if self.default_uom.category != self.purchase_uom.category:
+                    self.purchase_uom = self.default_uom
             else:
-                changes['purchase_uom'] = self.default_uom.id
-        return changes
+                self.purchase_uom = self.default_uom
+
+    @classmethod
+    def view_attributes(cls):
+        return super(Template, cls).view_attributes() + [
+            ('//page[@id="suppliers"]', 'states', {
+                    'invisible': ~Eval('purchasable'),
+                    })]
 
     @classmethod
     def write(cls, *args):
@@ -215,7 +222,7 @@ class ProductSupplier(ModelSQL, ModelView, MatchMixin):
     @staticmethod
     def order_sequence(tables):
         table, _ = tables[None]
-        return [table.sequence == None, table.sequence]
+        return [table.sequence == Null, table.sequence]
 
     @staticmethod
     def default_company():
@@ -231,9 +238,7 @@ class ProductSupplier(ModelSQL, ModelView, MatchMixin):
     @fields.depends('party')
     def on_change_party(self):
         cursor = Transaction().cursor
-        changes = {
-            'currency': self.default_currency(),
-            }
+        self.currency = self.default_currency()
         if self.party:
             table = self.__table__()
             cursor.execute(*table.select(table.currency,
@@ -242,15 +247,18 @@ class ProductSupplier(ModelSQL, ModelView, MatchMixin):
                     order_by=Count(Literal(1)).desc))
             row = cursor.fetchone()
             if row:
-                changes['currency'], = row
-        return changes
+                self.currency, = row
 
     def get_rec_name(self, name):
         return '%s @ %s' % (self.product.rec_name, self.party.rec_name)
 
     @classmethod
     def search_rec_name(cls, name, clause):
-        return ['OR',
+        if clause[1].startswith('!') or clause[1].startswith('not '):
+            bool_op = 'AND'
+        else:
+            bool_op = 'OR'
+        return [bool_op,
             ('product',) + tuple(clause[1:]),
             ('party',) + tuple(clause[1:]),
             ]
@@ -295,7 +303,8 @@ class ProductSupplierPrice(ModelSQL, ModelView, MatchMixin):
     product_supplier = fields.Many2One('purchase.product_supplier',
             'Supplier', required=True, ondelete='CASCADE')
     quantity = fields.Float('Quantity', required=True, help='Minimal quantity')
-    unit_price = fields.Numeric('Unit Price', required=True, digits=(16, 4))
+    unit_price = fields.Numeric('Unit Price', required=True,
+        digits=price_digits)
     sequence = fields.Integer('Sequence')
 
     @classmethod
@@ -322,7 +331,7 @@ class ProductSupplierPrice(ModelSQL, ModelView, MatchMixin):
     @staticmethod
     def order_sequence(tables):
         table, _ = tables[None]
-        return [table.sequence == None, table.sequence]
+        return [table.sequence == Null, table.sequence]
 
     @staticmethod
     def default_quantity():
