@@ -1,9 +1,10 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 import datetime
 from itertools import chain
 from decimal import Decimal
-from sql import Table, Literal
+
+from sql import Table, Literal, Null
 from sql.functions import Overlay, Position
 from sql.aggregate import Count
 from sql.operators import Concat
@@ -17,6 +18,9 @@ from trytond.pyson import Eval, Bool, If, PYSONEncoder, Id
 from trytond.transaction import Transaction
 from trytond.pool import Pool, PoolMeta
 
+from trytond.modules.account.tax import TaxableMixin
+from trytond.modules.product import price_digits
+
 __all__ = ['Purchase', 'PurchaseIgnoredInvoice',
     'PurchaseRecreadtedInvoice', 'PurchaseLine', 'PurchaseLineTax',
     'PurchaseLineIgnoredMove', 'PurchaseLineRecreatedMove', 'PurchaseReport',
@@ -31,7 +35,7 @@ _DEPENDS = ['state']
 _ZERO = Decimal(0)
 
 
-class Purchase(Workflow, ModelSQL, ModelView):
+class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
     'Purchase'
     __name__ = 'purchase.purchase'
     _rec_name = 'reference'
@@ -152,8 +156,10 @@ class Purchase(Workflow, ModelSQL, ModelView):
     @classmethod
     def __setup__(cls):
         super(Purchase, cls).__setup__()
-        cls._order.insert(0, ('purchase_date', 'DESC'))
-        cls._order.insert(1, ('id', 'DESC'))
+        cls._order = [
+            ('purchase_date', 'DESC'),
+            ('id', 'DESC'),
+            ]
         cls._error_messages.update({
                 'warehouse_required': ('A warehouse must be defined for '
                     'quotation of purchase "%s".'),
@@ -288,8 +294,8 @@ class Purchase(Workflow, ModelSQL, ModelView):
                                 purchase_line.id))
                         ).select(sql_table.id,
                             where=((sql_table.state == 'confirmed')
-                                & ((invoice_line.id != None)
-                                    | (move.id != None))))
+                                & ((invoice_line.id != Null)
+                                    | (move.id != Null))))
             cursor.execute(*sql_table.update(
                     columns=[sql_table.state],
                     values=['processing'],
@@ -355,23 +361,19 @@ class Purchase(Workflow, ModelSQL, ModelView):
     @fields.depends('party', 'payment_term', 'lines')
     def on_change_party(self):
         pool = Pool()
-        PaymentTerm = pool.get('account.invoice.payment_term')
         Currency = pool.get('currency.currency')
         cursor = Transaction().cursor
         table = self.__table__()
-        changes = {
-            'invoice_address': None,
-            'payment_term': None,
-            }
+        self.invoice_address = None
+        self.payment_term = self.default_payment_term()
         if not self.lines:
-            changes['currency'] = self.default_currency()
-            changes['currency_digits'] = self.default_currency_digits()
-        invoice_address = None
-        payment_term = None
+            self.currency = self.default_currency()
+            self.currency_digits = self.default_currency_digits()
+        self.invoice_address = None
         if self.party:
-            invoice_address = self.party.address_get(type='invoice')
+            self.invoice_address = self.party.address_get(type='invoice')
             if self.party.supplier_payment_term:
-                payment_term = self.party.supplier_payment_term
+                self.payment_term = self.party.supplier_payment_term
 
             if not self.lines:
                 subquery = table.select(table.currency,
@@ -384,24 +386,8 @@ class Purchase(Workflow, ModelSQL, ModelView):
                 row = cursor.fetchone()
                 if row:
                     currency_id, = row
-                    currency = Currency(currency_id)
-                    changes['currency'] = currency.id
-                    changes['currency_digits'] = currency.digits
-
-        if invoice_address:
-            changes['invoice_address'] = invoice_address.id
-            changes['invoice_address.rec_name'] = invoice_address.rec_name
-        else:
-            changes['invoice_address'] = None
-        if payment_term:
-            changes['payment_term'] = payment_term.id
-            changes['payment_term.rec_name'] = payment_term.rec_name
-        else:
-            changes['payment_term'] = self.default_payment_term()
-            if changes['payment_term']:
-                changes['payment_term.rec_name'] = PaymentTerm(
-                    changes['payment_term']).rec_name
-        return changes
+                    self.currency = Currency(currency_id)
+                    self.currency_digits = self.currency.digits
 
     @fields.depends('currency')
     def on_change_with_currency_digits(self, name=None):
@@ -417,7 +403,7 @@ class Purchase(Workflow, ModelSQL, ModelView):
                 return self.party.lang.code
         return Config.get_language()
 
-    def get_tax_context(self):
+    def _get_tax_context(self):
         context = {}
         if self.party and self.party.lang:
             context['language'] = self.party.lang.code
@@ -425,91 +411,47 @@ class Purchase(Workflow, ModelSQL, ModelView):
 
     @fields.depends('lines', 'currency', 'party')
     def on_change_lines(self):
-        pool = Pool()
-        Tax = pool.get('account.tax')
-        Invoice = pool.get('account.invoice')
-        Configuration = pool.get('account.configuration')
-
-        config = Configuration(1)
-
-        changes = {
-            'untaxed_amount': Decimal('0.0'),
-            'tax_amount': Decimal('0.0'),
-            'total_amount': Decimal('0.0'),
-            }
+        self.untaxed_amount = Decimal('0.0')
+        self.tax_amount = Decimal('0.0')
+        self.total_amount = Decimal('0.0')
+        taxes = {}
         if self.lines:
-            context = self.get_tax_context()
-            taxes = {}
-
-            def round_taxes():
-                if self.currency:
-                    for key, value in taxes.iteritems():
-                        taxes[key] = self.currency.round(value)
-
             for line in self.lines:
-                if getattr(line, 'type', 'line') != 'line':
-                    continue
-                changes['untaxed_amount'] += (getattr(line, 'amount', None)
-                    or Decimal(0))
+                self.untaxed_amount += getattr(line, 'amount', None) or 0
+            taxes = self._get_taxes()
+            self.tax_amount = sum(t['amount'] for t in taxes.itervalues())
+        if self.currency:
+            self.untaxed_amount = self.currency.round(self.untaxed_amount)
+            self.tax_amount = self.currency.round(self.tax_amount)
+        self.total_amount = self.untaxed_amount + self.tax_amount
+        if self.currency:
+            self.total_amount = self.currency.round(self.total_amount)
 
-                with Transaction().set_context(context):
-                    tax_list = Tax.compute(getattr(line, 'taxes', []),
-                        getattr(line, 'unit_price', None) or Decimal('0.0'),
-                        getattr(line, 'quantity', None) or 0.0)
-                for tax in tax_list:
-                    key, val = Invoice._compute_tax(tax, 'in_invoice')
-                    if key not in taxes:
-                        taxes[key] = val['amount']
-                    else:
-                        taxes[key] += val['amount']
-                if config.tax_rounding == 'line':
-                    round_taxes()
-            if config.tax_rounding == 'document':
-                round_taxes()
-            changes['tax_amount'] = sum(taxes.itervalues(), Decimal('0.0'))
-        if self.currency:
-            changes['untaxed_amount'] = self.currency.round(
-                changes['untaxed_amount'])
-            changes['tax_amount'] = self.currency.round(changes['tax_amount'])
-        changes['total_amount'] = (changes['untaxed_amount']
-            + changes['tax_amount'])
-        if self.currency:
-            changes['total_amount'] = self.currency.round(
-                changes['total_amount'])
-        return changes
+    @property
+    def taxable_lines(self):
+        taxable_lines = []
+        # In case we're called from an on_change we have to use some sensible
+        # defaults
+        for line in self.lines:
+            if getattr(line, 'type', None) != 'line':
+                continue
+            taxable_lines.append(tuple())
+            for attribute, default_value in (
+                    ('taxes', []),
+                    ('unit_price', Decimal(0)),
+                    ('quantity', 0.0)):
+                value = getattr(line, attribute, None)
+                taxable_lines[-1] += (value
+                    if value is not None else default_value,)
+        return taxable_lines
+
+    @property
+    def tax_type(self):
+        return 'invoice'
 
     def get_tax_amount(self):
-        pool = Pool()
-        Tax = pool.get('account.tax')
-        Invoice = pool.get('account.invoice')
-        Configuration = pool.get('account.configuration')
-
-        config = Configuration(1)
-
-        context = self.get_tax_context()
-        taxes = {}
-
-        def round_taxes():
-            for key, value in taxes.iteritems():
-                taxes[key] = self.currency.round(value)
-
-        for line in self.lines:
-            if line.type != 'line':
-                continue
-            with Transaction().set_context(context):
-                tax_list = Tax.compute(line.taxes, line.unit_price,
-                    line.quantity)
-            for tax in tax_list:
-                key, val = Invoice._compute_tax(tax, 'in_invoice')
-                if key not in taxes:
-                    taxes[key] = val['amount']
-                else:
-                    taxes[key] += val['amount']
-            if config.tax_rounding == 'line':
-                round_taxes()
-        if config.tax_rounding == 'document':
-            round_taxes()
-        return sum(taxes.itervalues(), _ZERO)
+        taxes = self._get_taxes().itervalues()
+        return sum(tax['amount'] for tax in taxes)
 
     @classmethod
     def get_amount(cls, purchases, names):
@@ -639,15 +581,23 @@ class Purchase(Workflow, ModelSQL, ModelView):
     @classmethod
     def search_rec_name(cls, name, clause):
         _, operator, value = clause
+        if operator[1].startswith('!') or operator[1].startswith('not '):
+            bool_op = 'AND'
+        else:
+            bool_op = 'OR'
         names = value.split(' - ', 1)
-        domain = ['OR',
+        domain = [bool_op,
             ('reference', operator, names[0]),
             ('supplier_reference', operator, names[0]),
             ]
         if len(names) != 1 and names[1]:
-            domain = [domain, ('party', operator, names[1])]
+            domain = [bool_op, domain, ('party', operator, names[1])]
         return domain
 
+    @classmethod
+    def view_attributes(cls):
+        return [('//field[@name="comment"]', 'spell', Eval('party_lang'))]
+
     @classmethod
     def copy(cls, purchases, default=None):
         if default is None:
@@ -956,7 +906,7 @@ class PurchaseLine(ModelSQL, ModelView):
     product_uom_category = fields.Function(
         fields.Many2One('product.uom.category', 'Product Uom Category'),
         'on_change_with_product_uom_category')
-    unit_price = fields.Numeric('Unit Price', digits=(16, 4),
+    unit_price = fields.Numeric('Unit Price', digits=price_digits,
         states={
             'invisible': Eval('type') != 'line',
             'required': Eval('type') == 'line',
@@ -973,9 +923,12 @@ class PurchaseLine(ModelSQL, ModelView):
     note = fields.Text('Note')
     taxes = fields.Many2Many('purchase.line-account.tax',
         'line', 'tax', 'Taxes',
+        order=[('tax.sequence', 'ASC'), ('tax.id', 'ASC')],
         domain=[('parent', '=', None), ['OR',
                 ('group', '=', None),
                 ('group.kind', 'in', ['purchase', 'both'])],
+                ('company', '=',
+                    Eval('_parent_purchase', {}).get('company', -1)),
             ],
         states={
             'invisible': Eval('type') != 'line',
@@ -1037,7 +990,7 @@ class PurchaseLine(ModelSQL, ModelView):
     @staticmethod
     def order_sequence(tables):
         table, _ = tables[None]
-        return [table.sequence == None, table.sequence]
+        return [table.sequence == Null, table.sequence]
 
     @staticmethod
     def default_type():
@@ -1098,6 +1051,7 @@ class PurchaseLine(ModelSQL, ModelView):
             context['uom'] = self.unit.id
         else:
             self.product.purchase_uom.id
+        context['taxes'] = [t.id for t in self.taxes]
         return context
 
     @fields.depends('product', 'unit', 'quantity', 'description',
@@ -1107,8 +1061,7 @@ class PurchaseLine(ModelSQL, ModelView):
         Product = Pool().get('product.product')
 
         if not self.product:
-            return {}
-        res = {}
+            return
 
         context = {}
         party = None
@@ -1117,69 +1070,70 @@ class PurchaseLine(ModelSQL, ModelView):
             if party.lang:
                 context['language'] = party.lang.code
 
-        category = self.product.purchase_uom.category
-        if not self.unit or self.unit not in category.uoms:
-            res['unit'] = self.product.purchase_uom.id
-            self.unit = self.product.purchase_uom
-            res['unit.rec_name'] = self.product.purchase_uom.rec_name
-            res['unit_digits'] = self.product.purchase_uom.digits
-
-        with Transaction().set_context(self._get_context_purchase_price()):
-            res['unit_price'] = Product.get_purchase_price([self.product],
-                abs(self.quantity or 0))[self.product.id]
-            if res['unit_price']:
-                res['unit_price'] = res['unit_price'].quantize(
-                    Decimal(1) / 10 ** self.__class__.unit_price.digits[1])
-        res['taxes'] = []
+        # Set taxes before unit_price to have taxes in context of purchase
+        # price
+        taxes = []
         pattern = self._get_tax_rule_pattern()
         for tax in self.product.supplier_taxes_used:
             if party and party.supplier_tax_rule:
                 tax_ids = party.supplier_tax_rule.apply(tax, pattern)
                 if tax_ids:
-                    res['taxes'].extend(tax_ids)
+                    taxes.extend(tax_ids)
                 continue
-            res['taxes'].append(tax.id)
+            taxes.append(tax)
         if party and party.supplier_tax_rule:
             tax_ids = party.supplier_tax_rule.apply(None, pattern)
             if tax_ids:
-                res['taxes'].extend(tax_ids)
+                taxes.extend(tax_ids)
+        self.taxes = taxes
+
+        category = self.product.purchase_uom.category
+        if not self.unit or self.unit not in category.uoms:
+            self.unit = self.product.purchase_uom
+            self.unit_digits = self.product.purchase_uom.digits
+
+        with Transaction().set_context(self._get_context_purchase_price()):
+            self.unit_price = Product.get_purchase_price([self.product],
+                abs(self.quantity or 0))[self.product.id]
+            if self.unit_price:
+                self.unit_price = self.unit_price.quantize(
+                    Decimal(1) / 10 ** self.__class__.unit_price.digits[1])
 
         if not self.description:
             with Transaction().set_context(context):
-                res['description'] = Product(self.product.id).rec_name
+                self.description = Product(self.product.id).rec_name
 
-        self.unit_price = res['unit_price']
         self.type = 'line'
-        res['amount'] = self.on_change_with_amount()
-        return res
+        self.amount = self.on_change_with_amount()
 
     @fields.depends('product')
     def on_change_with_product_uom_category(self, name=None):
         if self.product:
             return self.product.default_uom_category.id
 
-    @fields.depends('product', 'quantity', 'unit',
+    @fields.depends('product', 'quantity', 'unit', 'taxes',
         '_parent_purchase.currency', '_parent_purchase.party',
         '_parent_purchase.purchase_date')
     def on_change_quantity(self):
         Product = Pool().get('product.product')
 
         if not self.product:
-            return {}
-        res = {}
+            return
 
         with Transaction().set_context(self._get_context_purchase_price()):
-            res['unit_price'] = Product.get_purchase_price([self.product],
+            self.unit_price = Product.get_purchase_price([self.product],
                 abs(self.quantity or 0))[self.product.id]
-            if res['unit_price']:
-                res['unit_price'] = res['unit_price'].quantize(
+            if self.unit_price:
+                self.unit_price = self.unit_price.quantize(
                     Decimal(1) / 10 ** self.__class__.unit_price.digits[1])
-        return res
 
-    @fields.depends('product', 'quantity', 'unit',
-        '_parent_purchase.currency', '_parent_purchase.party')
+    @fields.depends(methods=['quantity'])
     def on_change_unit(self):
-        return self.on_change_quantity()
+        self.on_change_quantity()
+
+    @fields.depends(methods=['quantity'])
+    def on_change_taxes(self):
+        self.on_change_quantity()
 
     @fields.depends('type', 'quantity', 'unit_price', 'unit',
         '_parent_purchase.currency')
@@ -1289,10 +1243,17 @@ class PurchaseLine(ModelSQL, ModelView):
             if old_invoice_line.type != 'line':
                 continue
             if old_invoice_line.id not in skip_ids:
-                quantity -= Uom.compute_qty(old_invoice_line.unit,
-                        old_invoice_line.quantity, self.unit)
+                if old_invoice_line.invoice:
+                    old_invoice_type = old_invoice_line.invoice.type
+                else:
+                    old_invoice_type = old_invoice_line.invoice_type
 
-        rounding = self.unit.rounding if self.unit else 0
+                sign = 1 if old_invoice_type == invoice_type else -1
+
+                quantity -= Uom.compute_qty(old_invoice_line.unit,
+                        sign * old_invoice_line.quantity, self.unit)
+
+        rounding = self.unit.rounding if self.unit else 0.01
         invoice_line.quantity = Uom.round(quantity, rounding)
         if invoice_line.quantity <= 0:
             return []
@@ -1319,6 +1280,11 @@ class PurchaseLine(ModelSQL, ModelView):
                     {'purchase': self.purchase.rec_name})
         return [invoice_line]
 
+    @classmethod
+    def view_attributes(cls):
+        return [('//field[@name="note"]|//field[@name="description"]',
+                'spell', Eval('_parent_purchase', {}).get('party_lang'))]
+
     @classmethod
     def copy(cls, lines, default=None):
         if default is None:
@@ -1485,7 +1451,7 @@ class HandleShipmentException(Wizard):
     ask = StateView('purchase.handle.shipment.exception.ask',
         'purchase.handle_shipment_exception_ask_view_form', [
             Button('Cancel', 'end', 'tryton-cancel'),
-            Button('Ok', 'handle', 'tryton-ok', default=True),
+            Button('OK', 'handle', 'tryton-ok', default=True),
             ])
     handle = StateTransition()
 
@@ -1556,7 +1522,7 @@ class HandleInvoiceException(Wizard):
     ask = StateView('purchase.handle.invoice.exception.ask',
         'purchase.handle_invoice_exception_ask_view_form', [
             Button('Cancel', 'end', 'tryton-cancel'),
-            Button('Ok', 'handle', 'tryton-ok', default=True),
+            Button('OK', 'handle', 'tryton-ok', default=True),
             ])
     handle = StateTransition()
 
